///#include "complex.hh"
//#include <math.h>

#define GSL_DBL_EPSILON 2.2204460492503131e-16
#define GSL_SQRT_DBL_EPSILON   1.4901161193847656e-08
#define INT_THRESHOLD (1000.0*GSL_DBL_EPSILON)


/* Taken from

   Computation of Hypergeometric Functions
   by John Pearson

   2009, Oxford   */

dcomplex chf_1F1_series(dcomplex a, dcomplex b, dcomplex z) {
  // previous step
  dcomplex Ap = 1.;
  dcomplex Sp = Ap;
  // current step
  dcomplex A = z*a/b;
  dcomplex S = Sp + A;
  //
  int MaxIter = 500;
  double tolerance = 1.e-10;
  for(int k=1; k<MaxIter; k++) {
    Ap = A;
    A *= (a+k)/(b+k) * z/(k+1.);
    if( ABS(A)/ABS(S) < tolerance && ABS(Ap)/ABS(Sp) < tolerance) return S+A;
    Sp = S;
    S += A;
  }
  cout << "ERROR: 1F1 series did not converge" << endl;
  return S+A;
}

dcomplex chf_1F1_reg_series(dcomplex a, dcomplex b, dcomplex z) {
  double rintb = floor(real(b + 0.5));
  bool b_zero_or_neg_int = ( fabs(imag(b)) < INT_THRESHOLD && real(b) < 0.5 && fabs(real(b - rintb)) < INT_THRESHOLD );
  int first_k = 0;
  if(b_zero_or_neg_int) {
    first_k = 1 - (int)rintb;
    cout << "first_k  " << first_k << endl;
  }
  // previous step
  dcomplex Ap = Gamma(a+first_k)/Gamma(a)/Gamma(b+first_k) * pow(z,first_k) / gsl_sf_fact(first_k);
  dcomplex Sp = Ap;
  // current step
  dcomplex A = Ap * z/(first_k+1.) * (a+first_k)/(b+first_k);
  dcomplex S = Sp + A;
  //
  int MaxIter = 500;
  double tolerance = 1.e-10;
  for(int k=first_k+1; k<MaxIter+first_k; k++) {
    Ap = A;
    A *= (a+k)/(b+k) * z/(k+1.);
    if( ABS(A)/ABS(S) < tolerance && ABS(Ap)/ABS(Sp) < tolerance) return S+A;
    Sp = S;
    S += A;
  }
  cout << "ERROR: regularized 1F1 series did not converge" << endl;
  return S+A;
}



dcomplex chf_1F1_asympt(dcomplex a, dcomplex b, dcomplex z) {

  return 0;
}
dcomplex chf_1F1_reg_asympt(dcomplex a, dcomplex b, dcomplex z) {

  return 0;
}


dcomplex chf_1F1(dcomplex a, dcomplex b, dcomplex z) {
  dcomplex res;
  if(1) res = chf_1F1_series(a, b, z);
  else res = chf_1F1_asympt(a, b, z);
  return res;
}

dcomplex chf_1F1_reg(dcomplex a, dcomplex b, dcomplex z) {
  dcomplex res;
  if(1) res = chf_1F1_reg_series(a, b, z);
  else res = chf_1F1_reg_asympt(a, b, z);
  return res;
}


dcomplex chf_U(dcomplex a, dcomplex b, dcomplex z) {
  cout << "a="<< a << "    b=" << b << "    z=" << z << endl;
  //
  double rintb   = floor(real(b + 0.5));
  bool b_zero_or_neg_int = ( fabs(imag(b)) < INT_THRESHOLD && real(b) < 0.5 && fabs(real(b - rintb)) < INT_THRESHOLD );
  dcomplex M1, M2;
  if(b_zero_or_neg_int) M1 = chf_1F1_reg(a, b, z);
  else M1 = chf_1F1(a, b, z)/Gamma(b);
  //
  dcomplex b2 = 2.-b;
  rintb   = floor(real(b2 + 0.5));
  b_zero_or_neg_int = ( fabs(imag(b2)) < INT_THRESHOLD && real(b2) < 0.5 && fabs(real(b2 - rintb)) < INT_THRESHOLD );
  if(b_zero_or_neg_int) M2 = chf_1F1_reg(1.+a-b, 2.-b, z);
  else M2 = chf_1F1(1.+a-b, 2.-b, z)/Gamma(2.-b);
  //
  cout << "M1="<< M1 << "    M2=" << M2 << endl;
  return ( M1/Gamma(1.+a-b) - pow(z,1.-b)*M2/Gamma(a) ) * M_PI / sin(M_PI*b);
}

